"""
Handles data acquisition, loading and saving operations.
"""

import sys
import os
import pandas as pd
import handlers
import pickle

HANDLER_NAME = os.environ.get('FAST_DS_HANDLER', 'LocalFileHandler')
INPUT_DIR  = os.environ.get('FAST_DS_INPUT_DIR', 'data')
OUTPUT_DIR = os.environ.get('FAST_DS_OUTPUT_DIR', 'results')

HANDLER = getattr(handlers, HANDLER_NAME)


def read_fig(full_path_name):
    with HANDLER.fopen(full_path_name, 'rb') as image_content:
        return image_content.read()


def to_fig(fig, full_path_name, tight=True, remove_title=False):
    """
    to_fig(fig, fname, case_name='', dir_name=OUTPUT_DIR, tight=True, remove_title=False)

    Saves a matplotlib compatible plot in "dir_name/prefix/fname.pdf"
    Input parameters:
      - fig: instance returned by plt.figure().
      - fname: file name (without extensioon) where the image will be saved.
      - prefix: subdirectory under "dir_name" where the image file will be saved.
      - dir_name: base directory for saving the image.
      - tight: enforces matplotlib tight laypout.
      - remove_title: removes title and suptitle before saving the file (useful for papers and articles).
    """
    if tight: fig.tight_layout()
    if remove_title:
        fig.title(' ')
        fig.suptitle(' ')
    
    fname, _ = os.path.splitext(full_path_name)
    
    fmt_map = {
        'pdf' : {},
        'png' : dict(dpi=300),
    }
    
    for fmt, opt in fmt_map.items():
        full_name = f'{fname}.{fmt}'    
        with HANDLER.fopen(full_name , 'wb') as f:
            fig.savefig(f, format=fmt, **opt)


def read_pickle(full_path_name):
    with HANDLER.fopen(full_path_name, 'rb') as f:
        return pickle.load(f)

    
def to_pickle(obj, full_path_name):
    with HANDLER.fopen(full_path_name, 'wb') as f:
        pickle.dump(obj, f)
    

def to_excel(df, full_path_name, **kwargs):
    """
   Saves an Excel data file to "full_path_name".
    Input parameters:
      - full_path_name: The full path to the file to be generated.
      - **kwargs: to be passed to pandas.to_excel
    """
    with HANDLER.fopen(full_path_name, 'wb') as f:
        df.to_excel(f, **kwargs)


def get_function(pandas_ref, fname, operation):
    assert operation in ['read', 'to'], 'Operation must be either "read" or "to"'

    #We will employ the file name extension to derive the filetype we wish to save
    #and the function that should handle it.
    _, ext = os.path.splitext(fname)
    ext = ext[1:] # Removing the dot from the extension.

    format = None
    pandas_function = True
    if ext in ['xls', 'xlsx']:
        #If we are reading an excel file, we can use pd.read_excel.
        #Otherwise, we must use a custom version to handle S3 writting.
        pandas_function = operation == 'read'
        format = 'excel'
    elif ext in ['pdf', 'png', 'fig']:
        format = 'fig'
        pandas_function = False
    elif ext == 'pickle':
        format = 'pickle'
        pandas_function = False        
    else: format = ext

    #Here we derive whether we want to save or load in the provide file format.
    function_name = f'{operation}_{format}'

    #Getting a reference to the object from which the function we will call
    #to handle the oepration (pd.to_excel, to_fig, pf.read_parquet, etc.)
    obj_ref = pandas_ref if pandas_function else sys.modules[__name__]

    #Getting the function that will manage the content.
    func = getattr(obj_ref, function_name)

    #Getting the right wrapper. We need this wrapper to make sure
    #we will have the same interface between all saving possibilities.
    wrapper_func = save_with_pandas if pandas_function else save_without_pandas
    return func, wrapper_func


def save_with_pandas(df, full_name, func, **kwargs):
    func(full_name, **kwargs)


def save_without_pandas(df, full_name, func, **kwargs):
    func(df, full_name, **kwargs)


def save(df, fname, dir_name='', prefix='', **kwargs):
    func, wrapper_func = get_function(df, fname, 'to')
    full_name = HANDLER.get_full_path(dir_name, prefix, fname)
    wrapper_func(df, full_name, func, **kwargs)


def load(fname, dir_name='', prefix='', **kwargs):
    func, _ = get_function(pd, fname, 'read')
    file_obj = HANDLER.get_full_path(dir_name, prefix, fname)
    ret = func(file_obj, **kwargs)
    return ret



"""
Handles data pre_processing related operations.
"""

import re
import numpy as np


def cnpj_format(cnpj, ignore_null=False):
    """
    cnpj_format(cnpj, ignore_null=False)

    Returns a standartized (xx.xxx.xxx/xxxx-xx) version of the provided CNPJ.
    If the supplied CNPJ has less than 14 digits, this function will enforce
    it to have 14 digits by padding zeros to the left side of the CNPJ.
    
    If ignore_null is False (default), Null values will be converted to 00.000.000/0000-00.
    Otherwise, it will be returned as is (no change to it).
    """
    
    if ignore_null and cnpj in [None, np.nan]: return cnpj
    
    #Making sur that all CNPJ has only digits.
    non_digits = re.compile(r'\D')
    cnpj = non_digits.sub('', str(cnpj))

    #If the digits-only CNPJ has less than 14 digits, we complete it by
    #adding zeros to the left..
    digitos_faltando = 14 - len(cnpj)
    cnpj = '0'*digitos_faltando + cnpj

    #Formating it as "xx.xxx.xxx/xxxx-xx".
    filt = re.compile(r'(\d{2})(\d{3})(\d{3})(\d{4})(\d{2})')
    g = filt.match(cnpj)
    return f'{g.group(1)}.{g.group(2)}.{g.group(3)}/{g.group(4)}-{g.group(5)}'


def standard_nan(df, additional_nan_rep={}, std_nan_val=np.NaN, inplace=False):
    """
    Standartized multiple values that should be considered as null into a single value.
    """
    to_replace = {
        'na' : std_nan_val,
        'nan' : std_nan_val,
        '' : std_nan_val,
        '<NA>' : std_nan_val,
        'NÃO INFORMADO' : std_nan_val,
    }
    
    to_replace.update(additional_nan_rep)
    return df.replace(to_replace, inplace=inplace)


def drop_null_cols(df, threshold=0.6, inplace=False):
    """
    Remove columns which number of null samples is greater than the specified threshold.
    """
    null_freq = df.isna().sum() / len(df)
    df = df.drop(columns=null_freq.loc[null_freq > threshold].index, inplace=inplace)
    return df



def get_itb_range_averages(cat_val, is_integer=False):
    """
    Returns the mean value of ITB ranges for revenue, employees number, etc.
    """
    
    def get_number(val, is_integer=False):
        if not val: return np.NaN
        filt = re.compile(r'[\.,]')
        if is_integer: val = filt.sub('', val)
        return float(val)

    def get_order(val):
        orders_map = {
            'K' : 10**3,
            'M' : 10**6,
            'B' : 10**9,
        }
        return 1 if not val else orders_map[val[0]]

    filt = re.compile(r'(R\$)?\s*([\d\.,]+)\s*([K,M,I,B]*)\s*\-\s*(R\$)?\s*([\d\.,]+)\s*([K,M,I,B]*)\s*|\+\s*(R\$)?\s*([\d\.,]+)\s*\s*([K,M,I,B]*)')
    a = filt.match(cat_val)
    vec = np.array([
        get_number(a.group(2), is_integer) * get_order(a.group(3)),
        get_number(a.group(5), is_integer) * get_order(a.group(6)),
        get_number(a.group(8), is_integer) * get_order(a.group(9)),
        ])

    return np.nanmean(vec)